/*
 * Copyright 2002 (C) Michal Zawalich <michuz@pld.org.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <ghttp.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "cddb_handler.h"
#include "cdrom_handler.h"
#include "settings.h"

static gint cddb_sum(gint val);
static gchar *cddb_make_request(gchar * cmd, ghttp_request * request);
static GString *cddb_make_URI(gchar * cmd);
static gchar *cddb_read_line(gchar ** dataptr);
static gint cddb_genre_value(gchar * genre);
static void cddb_parse_title(gchar * buf, gchar * title, gchar * artist,
			     gchar * sep);
static void cddb_parse_title2(void);
static gchar *cddb_genre(gint genre);
static void cddb_process_line(gchar * inbuffer);

static char *cddb_genres[] = { "unknown", "blues", "classical", "country",
	"data", "folk", "jazz", "misc", "newage",
	"reggae", "rock", "soundtrack"
};
/********************************************************************/
static gint cddb_sum(gint val)
{
	char *bufptr, buf[16];
	int ret = 0;

	g_snprintf(buf, 16, "%lu", (unsigned long int) val);

	for (bufptr = buf; *bufptr != '\0'; bufptr++)
		ret += (*bufptr - '0');

	return ret;
}

/********************************************************************/
guint cddb_discid(void)
{
	gint i;
	gint tracksum = 0;
	gint discid;
	struct track_info *t;

	cdrom_get_stats();

	for (i = 0; i < g_slist_length(tracks_list); i++) {
		t = g_slist_nth_data(tracks_list, i);
		tracksum += cddb_sum(t->start_pos);
	}

	t = g_slist_nth_data(tracks_list, 0);
	discid = state.leadout_pos - t->start_pos;

	return (tracksum %
		0xFF) << 24 | discid << 8 | g_slist_length(tracks_list);
}

/********************************************************************/
/* query the cddb for the cd currently in the cdrom */
gboolean cddb_do_query(void)
{
	gint i;
	ghttp_request *request = NULL;
	GString *cmd;
	gchar *result;
	gchar *inbuffer;
	struct track_info *t;

	request = ghttp_request_new();

//  query->query_matches=0;

	cmd = g_string_new(NULL);
	g_string_sprintfa(cmd, "cddb+query+%08x+%d", cddb_discid(),
			  g_slist_length(tracks_list));

	for (i = 0; i < g_slist_length(tracks_list); i++) {
		t = g_slist_nth_data(tracks_list, i);
		g_string_sprintfa(cmd, "+%d", t->start_frame);
	}

	g_string_sprintfa(cmd, "+%d", state.leadout_pos);
	result = cddb_make_request(cmd->str, request);
	g_string_free(cmd, TRUE);

	if (!result) {
		ghttp_request_destroy(request);
		return FALSE;
	}

	inbuffer = cddb_read_line(&result);
	g_print("reply is [%s]\n", inbuffer);

	switch (strtol(strtok(inbuffer, " "), NULL, 10)) {
	case 200:		/* 200 - exact match */
		state.cddb_match = CDDB_MATCH_EXACT;
		state.cddb_list_genre =
		    cddb_genre_value(g_strstrip(strtok(NULL, " ")));
		sscanf(g_strstrip(strtok(NULL, " ")), "%xd",
		       &state.cddb_list_id);
		cddb_parse_title(g_strstrip(strtok(NULL, "")),
				 state.cddb_list_title,
				 state.cddb_list_artist, "/");
		break;
	case 210:		/* 210 - multiple exact matches */
		state.cddb_match = CDDB_MATCH_EXACT;
		while ((inbuffer = cddb_read_line(&result))) {
			state.cddb_list_genre =
			    cddb_genre_value(g_strstrip
					     (strtok(inbuffer, " ")));
			sscanf(g_strstrip(strtok(NULL, " ")), "%xd",
			       &state.cddb_list_id);
			cddb_parse_title(g_strstrip(strtok(NULL, "")),
					 state.cddb_list_title,
					 state.cddb_list_artist, "/");
		}
		break;
		/* 211 - inexact match */
//  case 211:
//    query->query_match=MATCH_INEXACT;
//    query->query_matches=0;

//    while((inbuffer=DiscDBReadLine(&result))) {
//      query->query_list[query->query_matches].list_genre=
//      DiscDBGenreValue(g_strstrip(strtok(inbuffer," ")));

//      sscanf(g_strstrip(strtok(NULL," ")),"%xd",
//           &query->query_list[query->query_matches].list_id);

//      DiscDBParseTitle(g_strstrip(strtok(NULL,"")),
//                     query->query_list[query->query_matches].list_title,
//                     query->query_list[query->query_matches].list_artist,
//                     "/");

//      query->query_matches++;
//    }

//    break;

	default:		/* no match */
		state.cddb_match = CDDB_MATCH_NOMATCH;
		ghttp_request_destroy(request);
		return (FALSE);
	}
	ghttp_request_destroy(request);
	return (TRUE);
}

/********************************************************************/
static gchar *cddb_make_request(gchar * cmd, ghttp_request * request)
{
	GString *uri;
	gchar user_agent[256];

	uri = cddb_make_URI(cmd);
	g_print("URI is %s\n", uri->str);

	ghttp_set_uri(request, uri->str);

	g_snprintf(user_agent, 256, "%s %s", state.cddb_program,
		   state.cddb_version);
	ghttp_set_header(request, "User-Agent", user_agent);
	ghttp_set_header(request, http_hdr_Connection, "close");
	ghttp_prepare(request);
	ghttp_process(request);

	g_string_free(uri, TRUE);

	return ghttp_get_body(request);
}

/********************************************************************/
static GString *cddb_make_URI(gchar * cmd)
{
	GString *uri = g_string_new(NULL);

	g_string_sprintf(uri,
			 "http://%s/%s?cmd=%s&hello=private+free.the.cddb+%s+%s&proto=%d",
			 state.cddb_server, state.cddb_cgi, cmd,
			 state.cddb_program, state.cddb_version, 5);
	return (uri);
}

/********************************************************************/
static gchar *cddb_read_line(gchar ** dataptr)
{
	char *data = *dataptr;
	char *pos;

	if (!data || !*data || *data == '.') {
		*dataptr = NULL;
		return (NULL);
	}

	for (pos = data; *pos; pos++) {
		if (*pos == '\n') {
			*pos = '\0';
			*dataptr = pos + 1;
			return (data);
		}
	}
	*dataptr = NULL;
	return (data);
}


/********************************************************************/
/* convert genre from text form into an integer value */
static gint cddb_genre_value(char *genre)
{
	gint pos;

	for (pos = 0; pos < 12; pos++)
		if (!strcmp(genre, cddb_genres[pos]))
			return (pos);

	return (0);
}

/********************************************************************/
/* Split string into title/artist */
static void cddb_parse_title(gchar * buf, gchar * title, gchar * artist,
			     gchar * sep)
{
	gchar *tmp = strtok(buf, sep);

	if (!tmp)
		return;

	g_snprintf(artist, 64, "%s", g_strstrip(tmp));
	tmp = strtok(NULL, "");

	if (tmp)
		g_snprintf(title, 64, "%s", g_strstrip(tmp));
	else
		strcpy(title, artist);
}

/********************************************************************/
/* read the actual cddb entry */
gboolean cddb_read(void)
{
	ghttp_request *request = ghttp_request_new();
	GString *cmd = g_string_new(NULL);
	gchar *result, *inbuffer;

	g_string_sprintf(cmd, "cddb+read+%s+%08x",
			 cddb_genre(state.cddb_list_genre),
			 state.cddb_list_id);
	result = cddb_make_request(cmd->str, request);
	g_string_free(cmd, TRUE);

	if (!result) {
		ghttp_request_destroy(request);
		return (FALSE);
	}

	inbuffer = cddb_read_line(&result);
	while ((inbuffer = cddb_read_line(&result)))
		cddb_process_line(inbuffer);

	/* both disc title and artist have been stuffed in the title field, so the
	   need to be separated */
	cddb_parse_title2();

	ghttp_request_destroy(request);
	return TRUE;
}

/********************************************************************/
static gchar *cddb_genre(gint genre)
{
	if (genre > 11)
		return ("unknown");

	return cddb_genres[genre];
}

/********************************************************************/
/* process a line of input data */
static void cddb_process_line(gchar * inbuffer)
{
	gint i;
	gint track;
	gchar *st;
	struct track_info *t;

	if (!strncasecmp(inbuffer, "DTITLE", 6)) {
		for (i = 0; i < g_slist_length(tracks_list); i++) {
			t = g_slist_nth_data(tracks_list, i);
			g_free(t->album);
			t->album = g_strdup(g_strstrip(g_strrstr(inbuffer + 7, "/") + 2));
		}
	} else if (!strncasecmp(inbuffer, "DYEAR", 5)) {
		strtok(inbuffer, "=");
		st = strtok(NULL, "");
		/* check if year is avaible */
		if (atoi(st) == 0)
			return;
		for (i = 0; i < g_slist_length(tracks_list); i++) {
			t = g_slist_nth_data(tracks_list, i);
			g_free(t->year);
			t->year = g_strstrip(g_strdup(st));
		}
	} else if (!strncasecmp(inbuffer, "DGENRE", 6)) {
		strtok(inbuffer, "=");
		st = strtok(NULL, "");
		if (st == NULL)
			return;

		st = g_strstrip(st);

		if (*st) {
			for (i = 0; i < g_slist_length(tracks_list); i++) {
				t = g_slist_nth_data(tracks_list, i);
				g_free(t->genre);
				t->genre = g_strdup(g_strstrip(st));
			}
		}
	} else if (!strncasecmp(inbuffer, "TARTIST", 7)) {
		track = atoi(strtok(inbuffer + 7, "="));
		t = g_slist_nth_data(tracks_list, track);
		st = strtok(NULL, "");
		if (st == NULL)
			return;
		g_free(t->artist);
		t->artist = g_strdup(g_strstrip(st));
	} else if (!strncasecmp(inbuffer, "TTITLE", 6)) {
		track = atoi(strtok(inbuffer + 6, "="));
		t = g_slist_nth_data(tracks_list, track);
		g_free(t->title);
		t->title = g_strdup(g_strstrip(strtok(NULL, "")));
	}
}

/********************************************************************/
/* Split string into title/artist */
static void cddb_parse_title2(void)
{
	struct track_info *t;
	gchar *tmp;
	gchar buf[128];
	gint i;

	for (i = 0; i < g_slist_length(tracks_list); i++) {
		t = g_slist_nth_data(tracks_list, i);
		strcpy(buf, t->title);
		if (g_strrstr(buf, "/") == NULL)
			return;

		tmp = strtok(buf, "/");
		if (!tmp)
			return;

		g_free(t->artist);
		t->artist = g_strdup_printf("%s", g_strstrip(tmp));

		tmp = strtok(NULL, "");
		if (tmp) {
			g_free(t->title);
			t->title = g_strdup_printf("%s", g_strstrip(tmp));
		} else {
			g_free(t->title);
			t->title = g_strdup(t->artist);
		}
	}
}
