#include "callbacks.h"
#include "cdrom_handler.h"
#include "error.h"
#include "settings.h"
#include <cdda_interface.h>
#include <cdda_paranoia.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <linux/cdrom.h>
#include <sys/ioctl.h>

#define OUTBUFSZ 32*1024

/* global 1-way track list */
GSList *tracks_list = NULL;

/* cdparanoia stuff */
static cdrom_drive *cd = NULL;
static cdrom_paranoia *p = NULL;

static long first_sector;	/* first sector of track */
static long last_sector;	/* last sector of track */
static time_t time_start;	/* remember when rip started */
static gint tr_num;		/* remember track number */

/* GLOBALS FOR BUFFERING CALLS */
static int bw_fd = -1;
static long bw_pos = 0;
static char bw_outbuf[OUTBUFSZ];

static gpointer pack_info(gint track, gint channels, gchar * time,
			  gint size);
static void put_num(long num, int f, int endianness, int bytes);
static void write_wav(int f, long bytes);
static void callback(long inpos, int function);
static int buffering_close(int fd);
static long buffering_write(int fd, char *buffer, long num);
static long blocking_write(int outf, char *buffer, long num);

/********************************************************************/
static gpointer pack_info(gint track, gint channels, gchar * time,
			  gint size)
{
	struct track_info *info = g_new(struct track_info, 1);

	info->track = track;
	info->wavsize = size;
	info->channels = channels;
	info->selected = TRUE;
	info->artist = g_strdup("unknown artist");
	info->title = g_strdup_printf("title%d", track);
	info->album = g_strdup("unknown album");
	info->year = g_strdup("1900");
	info->comment = g_strdup(" ");
	info->time = g_strdup(time);
	info->genre = g_strdup("unknown");

	return ((gpointer) info);
}

/********************************************************************/
gint cdrom_scan(void)
{
	gint i;
	gint channels;
	glong sec;
	glong off;
	glong audiolen = 0;
	gpointer info;
	gchar time[9];
	gint size;

	cd = cdda_identify(state.general_cdrom_path, CDDA_MESSAGE_PRINTIT, NULL);
	g_slist_free(tracks_list);

	/* cannot identify cd-rom device */
	if (cd == NULL) {
		gr_errno = GR_ENOCDROM;
		state.scan = FALSE;
		return (GR_ERROR);
	}

	cdda_verbose_set(cd, CDDA_MESSAGE_PRINTIT, CDDA_MESSAGE_PRINTIT);

	/* just taken from cdparanoia */
	switch (cdda_open(cd)) {
	case -2:
	case -3:
	case -4:
	case -5:
		gr_errno = GR_ENOAUDIO;
		state.scan = FALSE;
		return (GR_ERROR);
		break;
	case -6:
		gr_errno = GR_EFUCKDEV;
		state.scan = FALSE;
		return (GR_ERROR);
		break;
	case 0:
		break;
	default:
		gr_errno = GR_GENERRDEV;
		state.scan = FALSE;
		return (GR_ERROR);
	}

	cdda_open(cd);

	/* getting informations */
	for (i = 1; i <= cd->tracks; i++) {
		if (cdda_track_audiop(cd, i) != 0) {

			sec = cdda_track_firstsector(cd, i);
			off = cdda_track_lastsector(cd, i) - sec + 1;
			channels = cdda_track_channels(cd, i);
			size =
			    44 + (cdda_track_lastsector(cd, i) -
				  cdda_track_firstsector(cd,
							 i) + 1) * 2352;

			sprintf(time, "%02d:%02d:%02d",
				(int) (off / (60 * 75)),
				(int) ((off / 75) % 60), (int) (off % 75));

			info = pack_info(i, channels, time, size);
			tracks_list = g_slist_append(tracks_list, info);

			audiolen += off;
		}
	}

	/* appbar summary */
	barinfo =
	    g_strdup_printf("total time: %02d:%02d.%02d",
			    (int) (audiolen / (60 * 75)),
			    (int) ((audiolen / 75) % 60),
			    (int) (audiolen % 75));

	cdda_close(cd);
	cd = NULL;
	state.scan = TRUE;
	return (GR_SUCCESS);
}

/********************************************************************/
void *cdrom_get_track(gpointer data)
{
	struct track_info *track =
	    g_slist_nth_data(tracks_list, GPOINTER_TO_INT(data));
	long cursor;		/* cd read position */
	gint out;		/* file descryptor */

	/* lock thread mutex */
	g_mutex_lock(state.cdparanoia_mutex);

	if (state.progress_state == PROGRESS_ABORT)
		return (NULL);

	/* check cdrom */
	cd = cdda_identify(state.general_cdrom_path, CDDA_MESSAGE_PRINTIT, NULL);

	/* cannot identify cd-rom device */
	if (cd == NULL) {
		gr_errno = GR_ENOCDROM;
		state.scan = FALSE;
	}

	cdda_verbose_set(cd, CDDA_MESSAGE_PRINTIT, CDDA_MESSAGE_PRINTIT);

	/* just taken from cdparanoia */
	switch (cdda_open(cd)) {
	case -2:
	case -3:
	case -4:
	case -5:
		gr_errno = GR_ENOAUDIO;
		state.scan = FALSE;
		return (NULL);
		break;
	case -6:
		gr_errno = GR_EFUCKDEV;
		state.scan = FALSE;
		return (NULL);
		break;
	case 0:
		break;
	default:
		gr_errno = GR_GENERRDEV;
		state.scan = FALSE;
		return (NULL);
	}

	cdda_open(cd);

	/* get track info */
	first_sector = cdda_track_firstsector(cd, track->track);
	last_sector = cdda_track_lastsector(cd, track->track);

	/* cdparanoia init */
	p = paranoia_init(cd);
	paranoia_modeset(p, PARANOIA_MODE_FULL ^ PARANOIA_MODE_NEVERSKIP);
	paranoia_seek(p, cursor = first_sector, SEEK_SET);

	/* progress stuff */
	tr_num = track->track;

	/* time */
	time_start = time(NULL);

	/* create wav file */
	out = open(track->savefile, O_RDWR | O_CREAT | O_TRUNC, 0666);
	write_wav(out,
		  (last_sector - first_sector + 1) * CD_FRAMESIZE_RAW);

	/* main loop */
	while (cursor <= last_sector) {

		if (state.progress_state == PROGRESS_ABORT)
			break;

		/* read a sector */
		int16_t *readbuf = paranoia_read_limited(p, callback, 20);
		cursor++;
		callback(cursor * (CD_FRAMEWORDS) - 1, -2);
		buffering_write(out, ((char *) readbuf), CD_FRAMESIZE_RAW);
	}
	buffering_close(out);
	state.ripsize += track->wavsize;
	callback(cursor * (CD_FRAMESIZE_RAW / 2) - 1, -1);

	/* cleanup */
	if (state.progress_state == PROGRESS_ABORT)
		unlink(track->savefile);
	paranoia_free(p);
	cdda_close(cd);
	cd = NULL;
	p = NULL;

	/* unlock thread mutex */
	g_mutex_unlock(state.cdparanoia_mutex);
	pthread_exit(0);
	return (NULL);
}

/********************************************************************/
static void put_num(long num, int f, int endianness, int bytes)
{
	int i;
	unsigned char c;

	if (!endianness)
		i = 0;
	else
		i = bytes - 1;
	while (bytes--) {
		c = (num >> (i << 3)) & 0xff;
		if (write(f, &c, 1) == -1) {
			perror("Could not write to output.");
			exit(1);
		}
		if (endianness)
			i--;
		else
			i++;
	}
}

/********************************************************************/
static void write_wav(int f, long bytes)
{
	/* quick and dirty */

	write(f, "RIFF", 4);	/*  0-3 */
	put_num(bytes + 44 - 8, f, 0, 4);	/*  4-7 */
	write(f, "WAVEfmt ", 8);	/*  8-15 */
	put_num(16, f, 0, 4);	/* 16-19 */
	put_num(1, f, 0, 2);	/* 20-21 */
	put_num(2, f, 0, 2);	/* 22-23 */
	put_num(44100, f, 0, 4);	/* 24-27 */
	put_num(44100 * 2 * 2, f, 0, 4);	/* 28-31 */
	put_num(4, f, 0, 2);	/* 32-33 */
	put_num(16, f, 0, 2);	/* 34-35 */
	write(f, "data", 4);	/* 36-39 */
	put_num(bytes, f, 0, 4);	/* 40-43 */
}

/********************************************************************/
static void callback(long inpos, int function)
{
	long sector = inpos / CD_FRAMEWORDS;
	time_t now = time(NULL);
	gfloat size;
	gfloat speed;
	gfloat elapsed = (gfloat) now - (gfloat) time_start;
	struct track_info *track =
	    g_slist_nth_data(tracks_list, tr_num - 1);

	if (function == 9)
		return;

	/* 1x is 44100*2*2 = 176400 bytes/sec */
	if (elapsed < 0.1f)	/* 1/10 sec. */
		speed = 0.0f;	/* Avoid divide-by-0 at start */
	else
		speed =
		    (gfloat) (44 +
			      (sector - first_sector +
			       1) * 2352) / (176400.0f * elapsed);

	/* startup */
	if (speed >= 50.0f) {
		speed = 0.0f;
		time_start = now;
	}

	g_free(state.progress_text);
	g_free(state.progress_label1);
	g_free(state.progress_label2);
	state.progress_text = g_strdup_printf("x %0.1f", speed);
	state.progress_label1 = g_strdup_printf("INPUT: track %d", tr_num);
	state.progress_label2 =
	    g_strdup_printf("OUTPUT: %s",
			    g_strrstr(track->savefile, "/") + 1);
	state.progress =
	    ((float) (sector - first_sector) /
	     (last_sector - first_sector)) * 100;
	size = state.ripsize + 44 + (sector - first_sector + 1) * 2352;
	state.progress2 = (size / (float) state.totalsize) * 100;
}

/********************************************************************/
/* buffering_write() - buffers data to a specified size before writing.
 *
 * Restrictions:
 * - MUST CALL BUFFERING_CLOSE() WHEN FINISHED!!!
 */
static long buffering_write(int fd, char *buffer, long num)
{
	if (fd != bw_fd) {
		/* clean up after buffering for some other file */
		if (bw_fd >= 0 && bw_pos > 0) {
			if (blocking_write(bw_fd, bw_outbuf, bw_pos)) {
				perror
				    ("write (in buffering_write, flushing)");
			}
		}
		bw_fd = fd;
		bw_pos = 0;
	}

	if (bw_pos + num > OUTBUFSZ) {
		/* fill our buffer first, then write, then modify buffer and num */
		memcpy(&bw_outbuf[bw_pos], buffer, OUTBUFSZ - bw_pos);
		if (blocking_write(fd, bw_outbuf, OUTBUFSZ)) {
			perror("write (in buffering_write, full buffer)");
			return (-1);
		}
		num -= (OUTBUFSZ - bw_pos);
		buffer += (OUTBUFSZ - bw_pos);
		bw_pos = 0;
	}
	/* save data */
	memcpy(&bw_outbuf[bw_pos], buffer, num);
	bw_pos += num;

	return (0);
}

/********************************************************************/
/* buffering_close() - writes out remaining buffered data before closing
 * file.
 */
static int buffering_close(int fd)
{
	if (fd == bw_fd && bw_pos > 0) {
		/* write out remaining data and clean up */
		if (blocking_write(fd, bw_outbuf, bw_pos)) {
			perror("write (in buffering_close)");
		}
		bw_fd = -1;
		bw_pos = 0;
	}
	return (close(fd));
}

/********************************************************************/
static long blocking_write(int outf, char *buffer, long num)
{
	long words = 0, temp;

	while (words < num) {
		temp = write(outf, buffer + words, num - words);
		if (temp == -1) {
			if (errno != EINTR && errno != EAGAIN)
				return (-1);
			temp = 0;
		}
		words += temp;
	}
	return (0);
}

/********************************************************************/
void cdrom_get_stats(void)
{
	struct cdrom_tocentry te;
	gint fd;
	gint i;
	struct track_info *track;

	fd = open(state.general_cdrom_path, O_RDONLY | O_NONBLOCK);

	/* read the table of contents */
	for (i = 1; i <= g_slist_length(tracks_list); i++) {
		/* get info from cd */
		te.cdte_track = i;
		te.cdte_format = CDROM_MSF;
		ioctl(fd, CDROMREADTOCENTRY, &te);

		/* save start position */
		track = g_slist_nth_data(tracks_list, i - 1);
		track->start_pos =
		    te.cdte_addr.msf.minute * 60 + te.cdte_addr.msf.second;
		track->start_frame =
		    track->start_pos * 75 + te.cdte_addr.msf.frame;
	}

	/* save leadout */
	te.cdte_track = CDROM_LEADOUT;
	te.cdte_format = CDROM_MSF;
	ioctl(fd, CDROMREADTOCENTRY, &te);
	state.leadout_pos =
	    te.cdte_addr.msf.minute * 60 + te.cdte_addr.msf.second;
}
