/*
 * Copyright 2002 (C) Michal Zawalich <michuz@pld.org.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <gnome.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "error.h"
#include "interface.h"
#include "settings.h"
#include "userdir.h"

#define BUFSIZE 80

/* global state var */
state_t state;

/* tree_item_t structure */
typedef struct tree_item tree_item_t;
struct tree_item {
	const gchar *label;
	GtkWidget *frame;
	tree_item_t *child;
};

enum {
	SETTINGS_NAME_COLUMN,
	SETTINGS_FRAME_COLUMN,
	SETTINGS_NUM_COLUMNS
};

/* tree data */
static tree_item_t codecs[] = {
	{"mp3 lame", NULL, NULL},
	{"ogg vorbis", NULL, NULL},
	{"aac faac", NULL, NULL},
	{NULL}
};

static tree_item_t toplevel[] = {
	{"general", NULL, NULL},
	{"CDDB", NULL, NULL},
	{"codecs", NULL, codecs},
	{NULL}
};

/* widgets with data */
typedef struct {
	GtkWidget *active_frame; /* pointer to active frame */
	GtkWidget *hbox;         /* container for active frame */

	GtkWidget *general_cdrom_path_entry;
	GtkWidget *general_name_format_entry;
	
	GtkWidget *vorbis_quality_entry;
	GtkWidget *vorbis_mono_button;
	GtkWidget *vorbis_stereo_button;
} entry_t;

static entry_t entry;

static void selection_cb(GtkTreeSelection *selection, GtkTreeModel *model);
static void response_cb(GtkDialog *dialog, gint id, gpointer data);

static void vorbis_selection_changed_cb(GtkList *list, gpointer data);
static void vorbis_mono_toggled_cb(GtkToggleButton *togglebutton, gpointer data);
static void vorbis_stereo_toggled_cb(GtkToggleButton *togglebutton, gpointer data);

static GtkTreeModel *create_model(void);
static void create_general_frame(void);
static void create_cddb_frame(void);
static void create_lame_frame(void);
static void create_vorbis_frame(void);
static void create_faac_frame(void);
static void pack_data(void);
static void load_data(void);
/********************************************************************/
void settings_set_default(void)
{
	state.general_cdrom_path = g_strdup("/dev/cdrom");
	state.general_name_format = g_strdup_printf("%%p - %%t");
	
	state.vorbis_quality = 50;
	state.vorbis_channels = 2;
}

/********************************************************************/
void settings_load(void)
{
	FILE *file;
	gchar buf[BUFSIZE];
	gchar *tmp;
	struct cfv data;

	data.variable = "CONFIG_DIR"; /* system var. contains config directory name */
    data.home_dir = NULL;         /* autodetection */
    data.home_scd = NULL;         /* if something went bad - fuck it */
    data.subname = PACKAGE;       /* core filename */
    data.prefix = ".";        	/* if $CONFIG_DIR is not supported then add . */
    data.suffix = "rc";           /* if $CONFIG_DIR is not supported then add rc */
    data.mode = M_REGULAR_FILE;   /* file search */

	if ((file = maynullfopencfv (&data, "r")) == NULL)
		return;
	else {
		while (fgets(buf, BUFSIZE, file) != NULL) {
			/* cdrom device */
			if (strstr(buf, "general_cdrom_path=") != NULL) {
				tmp = strchr(buf, '=');
				g_free(state.general_cdrom_path);
				state.general_cdrom_path = g_strdup_printf("%s", tmp + 1);
				state.general_cdrom_path[strlen(state.general_cdrom_path) - 1] = '\0';
			} else if (strstr(buf, "general_name_format=") != NULL) {
				tmp = strchr(buf, '=');
				g_free(state.general_name_format);
				state.general_name_format = g_strdup_printf("%s", tmp + 1);
				state.general_name_format[strlen(state.general_name_format) - 1] = '\0';
			} else if (strstr(buf, "vorbis_quality=") != NULL) {
				tmp = strchr(buf, '=');
				state.vorbis_quality = atoi(tmp + 1);
			} else if (strstr(buf, "vorbis_channels=") != NULL) {
				tmp = strchr(buf, '=');
				state.vorbis_channels = atoi(tmp + 1);
			}
		}
		fclose(file);
	}
}

/********************************************************************/
void settings_save(void)
{
	FILE *file;
	gchar buf[BUFSIZE];
	struct cfv data;

	data.variable = "CONFIG_DIR"; /* system var. contains config directory name */
    data.home_dir = NULL;         /* autodetection */
    data.home_scd = NULL;         /* if something went bad - fuck it */
    data.subname = PACKAGE;       /* core filename */
    data.prefix = ".";        	/* if $CONFIG_DIR is not supported then add . */
    data.suffix = "rc";           /* if $CONFIG_DIR is not supported then add rc */
    data.mode = M_REGULAR_FILE;   /* file search */

	if ((file = notnullfopencfv(&data, "w")) == NULL) {
		gr_errno = GR_ESAVESETTINGS;
		gr_error();
		return;
	} else {
		sprintf(buf, "general_cdrom_path=%s\n", state.general_cdrom_path);
		fputs(buf, file);
		sprintf(buf, "general_name_format=%s\n", state.general_name_format);
		fputs(buf, file);
		
		sprintf(buf, "vorbis_quality=%d\n", state.vorbis_quality);
		fputs(buf, file);
		sprintf(buf, "vorbis_channels=%d\n", state.vorbis_channels);
		fputs(buf, file);
	}
	fclose(file);
}

/********************************************************************/
void create_settings_window(void)
{
	GtkWidget *window;
	GtkTreeModel *model;
	GtkWidget *tree_view;
	GtkWidget *vbox;
	GtkTreeSelection *selection;
	GtkCellRenderer *cell;
	GtkTreeViewColumn *column;
	
	/* frame constructors */
	create_general_frame();
	create_cddb_frame();
	create_lame_frame();
	create_vorbis_frame();
	create_faac_frame();
	
	window = gtk_dialog_new_with_buttons(PACKAGE " settings", GTK_WINDOW(application), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_APPLY, GTK_RESPONSE_APPLY, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	vbox = GTK_DIALOG(window)->vbox;
	entry.hbox = gtk_hbox_new(FALSE, BORDER_SMALL);

	/* tree */
	model = create_model();
	tree_view = gtk_tree_view_new();
	gtk_tree_view_set_model(GTK_TREE_VIEW (tree_view), GTK_TREE_MODEL (model));
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
	cell = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(NULL, cell, "text", SETTINGS_NAME_COLUMN, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW (tree_view), GTK_TREE_VIEW_COLUMN (column));
	
	/* packing */
	gtk_box_pack_start(GTK_BOX(vbox), entry.hbox, TRUE, TRUE, BORDER_NONE);
	gtk_box_pack_start(GTK_BOX(entry.hbox), tree_view, FALSE, TRUE, BORDER_NONE);

	/* other options */
	gtk_widget_set_size_request(window, 500, 400);
	gtk_widget_set_size_request(tree_view, 150, -1);
	gtk_tree_view_expand_all(GTK_TREE_VIEW(tree_view));
	gtk_tree_selection_set_mode(GTK_TREE_SELECTION(selection), GTK_SELECTION_BROWSE);
	pack_data();

	/* events */
	g_signal_connect(selection, "changed", G_CALLBACK(selection_cb), model);
	g_signal_connect(window, "response", G_CALLBACK(response_cb), window);
	
	/* showing widgets */
	gtk_widget_show_all(window);
}

/********************************************************************/
static GtkTreeModel *create_model(void)
{
	GtkTreeStore *model;
	GtkTreeIter iter;
	tree_item_t *i = toplevel;

	/* create tree store */
	model = gtk_tree_store_new(SETTINGS_NUM_COLUMNS, G_TYPE_STRING, G_TYPE_POINTER);
	
	/* add data to the tree store */
	while (i->label) {
		tree_item_t *j = i->child;
		
		gtk_tree_store_append(model, &iter, NULL);
		gtk_tree_store_set(model, &iter, SETTINGS_NAME_COLUMN, i->label, SETTINGS_FRAME_COLUMN, i->frame, -1);

		i++;
		if (!j) continue;
			
		/* add children */
		while(j->label) {
			GtkTreeIter child_iter;

			gtk_tree_store_append(model, &child_iter, &iter);
			gtk_tree_store_set(model, &child_iter, SETTINGS_NAME_COLUMN, j->label, SETTINGS_FRAME_COLUMN, j->frame, -1);
			j++;
		}
    }
	return (GTK_TREE_MODEL(model));
}

/********************************************************************/
static void pack_data(void)
{
	/* general */
	gtk_entry_set_text(GTK_ENTRY(entry.general_cdrom_path_entry), state.general_cdrom_path);
	gtk_entry_set_text(GTK_ENTRY(entry.general_name_format_entry), state.general_name_format);
	
	/* vorbis */
	gtk_entry_set_text(GTK_ENTRY(entry.vorbis_quality_entry), g_strdup_printf("%d", state.vorbis_quality));
	switch (state.vorbis_channels) {
	case 1:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(entry.vorbis_mono_button), TRUE);
		break;
	case 2:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(entry.vorbis_stereo_button), TRUE);
		break;
	}
}

/********************************************************************/
static void load_data(void)
{
	g_free(state.general_cdrom_path);
	g_free(state.general_name_format);

	state.general_cdrom_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry.general_cdrom_path_entry)));
	state.general_name_format = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry.general_name_format_entry)));
	
	state.vorbis_quality = atoi(gtk_entry_get_text(GTK_ENTRY(entry.vorbis_quality_entry)));
}

/********************************************************************/
static void selection_cb(GtkTreeSelection *selection, GtkTreeModel *model)
{
	GtkTreeIter iter;
	GtkWidget *frame;
	tree_item_t *codec = &toplevel[2]; /* set item to codec position */

	/* check if something is selected */
	if (!gtk_tree_selection_get_selected(selection, NULL, &iter))
		return;

	gtk_tree_model_get(model, &iter, SETTINGS_FRAME_COLUMN, &frame, -1);

	/* check if this is the same object */
	if (entry.active_frame == frame)
		return;

	/* exit if codecs is selected */
	if (frame == codec->frame)
		return;

	/* do not remove empty widget */
	if (entry.active_frame != NULL)
		gtk_container_remove(GTK_CONTAINER(entry.hbox), entry.active_frame);
	
	entry.active_frame = frame;
	gtk_box_pack_start(GTK_BOX(entry.hbox), entry.active_frame, TRUE, TRUE, BORDER_NONE);
}

/********************************************************************/
static void response_cb(GtkDialog *dialog, gint id, gpointer data)
{
	switch (id) {
	case GTK_RESPONSE_APPLY:
		load_data();
		settings_save();
		break;
	case GTK_RESPONSE_ACCEPT:
		load_data();
		settings_save();
	case GTK_RESPONSE_REJECT:
		gtk_widget_destroy(GTK_WIDGET(data));
		break;
	}
}

/********************************************************************/
static void create_general_frame(void)
{
	tree_item_t *item = &toplevel[0];
	GtkWidget *cdrom_label;
	GtkWidget *format_label;
	GtkWidget *format_table;
	GtkWidget *vbox;
	GtkWidget *separator1;
	GtkWidget *format_artist_label;
	GtkWidget *format_title_label;
	GtkWidget *format_album_label;
	GtkWidget *format_track_label;
	GtkWidget *format_year_label;
	GtkWidget *format_comment_label;
	
	item->frame = gtk_frame_new(item->label);
	vbox = gtk_vbox_new(FALSE, BORDER_NONE);
	cdrom_label = gtk_label_new("cd-rom device:");
	format_label = gtk_label_new("output format:");	
	entry.general_cdrom_path_entry = gtk_entry_new();
	entry.general_name_format_entry = gtk_entry_new();
	separator1 = gtk_hseparator_new();
	format_table = gtk_table_new(2, 3, FALSE);
	format_artist_label = gtk_label_new("%p - artist");
	format_title_label = gtk_label_new("%t - title");
	format_album_label = gtk_label_new("%a - album");
	format_track_label = gtk_label_new("%n - track");
	format_year_label = gtk_label_new("%y - year");
	format_comment_label = gtk_label_new("%c - comment");

	gtk_container_add(GTK_CONTAINER(item->frame), vbox);
	gtk_box_pack_start(GTK_BOX(vbox), cdrom_label, FALSE, FALSE, BORDER_SMALL);
	gtk_box_pack_start(GTK_BOX(vbox), entry.general_cdrom_path_entry, FALSE, FALSE, BORDER_NONE);
	gtk_box_pack_start(GTK_BOX(vbox), separator1, FALSE, FALSE, BORDER_SMALL);
	gtk_box_pack_start(GTK_BOX(vbox), format_label, FALSE, FALSE, BORDER_SMALL);
	gtk_box_pack_start(GTK_BOX(vbox), entry.general_name_format_entry, FALSE, FALSE, BORDER_NONE);
	gtk_box_pack_start(GTK_BOX(vbox), format_table, FALSE, FALSE, BORDER_SMALL);
	gtk_table_attach_defaults(GTK_TABLE(format_table), format_artist_label, 0, 1, 0, 1);
	gtk_table_attach_defaults(GTK_TABLE(format_table), format_title_label, 1, 2, 0, 1);
	gtk_table_attach_defaults(GTK_TABLE(format_table), format_album_label, 0, 1, 1, 2);
	gtk_table_attach_defaults(GTK_TABLE(format_table), format_year_label, 1, 2, 1, 2);
	gtk_table_attach_defaults(GTK_TABLE(format_table), format_track_label, 0, 1, 2, 3);
	gtk_table_attach_defaults(GTK_TABLE(format_table), format_comment_label, 1, 2, 2, 3);
	
	gtk_container_set_border_width(GTK_CONTAINER(vbox), BORDER_BIG);
	gtk_misc_set_alignment(GTK_MISC(cdrom_label), 0, 1);
	gtk_misc_set_alignment(GTK_MISC(format_label), 0, 1);
	gtk_misc_set_alignment(GTK_MISC(format_artist_label), 0, 0.5);
	gtk_misc_set_alignment(GTK_MISC(format_title_label), 0, 0.5);
	gtk_misc_set_alignment(GTK_MISC(format_album_label), 0, 0.5);
	gtk_misc_set_alignment(GTK_MISC(format_year_label), 0, 0.5);
	gtk_misc_set_alignment(GTK_MISC(format_track_label), 0, 0.5);
	gtk_misc_set_alignment(GTK_MISC(format_comment_label), 0, 0.5);
	g_object_ref(item->frame); /* for multiple use */

	gtk_widget_show_all(item->frame);
}

/********************************************************************/
static void create_cddb_frame(void)
{
	tree_item_t *item;
	
	item = &toplevel[1];
	item->frame = gtk_frame_new(item->label);
	g_object_ref(item->frame); /* for multiple use */
	gtk_widget_show_all(item->frame);
}

/********************************************************************/
static void create_lame_frame(void)
{
	tree_item_t *item;
	
	item = &codecs[0];
	item->frame = gtk_frame_new(item->label);
	g_object_ref(item->frame); /* for multiple use */
	gtk_widget_show_all(item->frame);
}

/********************************************************************/
static void create_vorbis_frame(void)
{
	tree_item_t *item = &codecs[1];
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *table;
	GtkWidget *bitrate_entry;
	GtkWidget *channels_label;
	GtkWidget *quality_label;
	GtkWidget *quality_combo;
	GtkWidget *combo_list;
	GList *items = NULL;
	
	items = g_list_append(items, "5");
	items = g_list_append(items, "10");
	items = g_list_append(items, "15");
	items = g_list_append(items, "20");
	items = g_list_append(items, "25");
	items = g_list_append(items, "30");
	items = g_list_append(items, "35");
	items = g_list_append(items, "40");
	items = g_list_append(items, "45");
	items = g_list_append(items, "50");
	items = g_list_append(items, "55");
	items = g_list_append(items, "60");
	items = g_list_append(items, "65");
	items = g_list_append(items, "70");
	items = g_list_append(items, "75");
	items = g_list_append(items, "80");
	items = g_list_append(items, "85");
	items = g_list_append(items, "90");
	items = g_list_append(items, "95");
	items = g_list_append(items, "100");
	
	item->frame = gtk_frame_new(item->label);
	vbox = gtk_vbox_new(FALSE, 0);
	hbox = gtk_hbox_new(FALSE, 0);
	table = gtk_table_new(3, 1, FALSE);
	channels_label = gtk_label_new("Channels");
	entry.vorbis_mono_button = gtk_radio_button_new_with_label(NULL, "Mono");
	entry.vorbis_stereo_button = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(entry.vorbis_mono_button), "Stereo");
	quality_label = gtk_label_new("Quality");
	bitrate_entry = gtk_entry_new_with_max_length(8);
	quality_combo = gtk_combo_new();
	entry.vorbis_quality_entry = GTK_COMBO(quality_combo)->entry;
	combo_list = GTK_COMBO(quality_combo)->list;

	gtk_container_add(GTK_CONTAINER(item->frame), vbox);
	gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, BORDER_SMALL);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, BORDER_SMALL);
	gtk_table_attach_defaults(GTK_TABLE(table), channels_label, 0, 1, 0, 1);
	gtk_table_attach_defaults(GTK_TABLE(table), entry.vorbis_mono_button, 1, 2, 0, 1);
	gtk_table_attach_defaults(GTK_TABLE(table), entry.vorbis_stereo_button, 2, 3, 0, 1);
	gtk_box_pack_start(GTK_BOX(hbox), quality_label, FALSE, FALSE, BORDER_NONE);
	gtk_box_pack_start(GTK_BOX(hbox), bitrate_entry, FALSE, FALSE, BORDER_BIG);
	gtk_box_pack_start(GTK_BOX(hbox), quality_combo, FALSE, FALSE, BORDER_BIG);
	
	gtk_container_set_border_width(GTK_CONTAINER(vbox), BORDER_BIG);
	gtk_combo_set_popdown_strings(GTK_COMBO(quality_combo), items);
	gtk_misc_set_alignment(GTK_MISC(channels_label), 0, 0.5);
	gtk_misc_set_alignment(GTK_MISC(quality_label), 0, 0.5);
	gtk_entry_set_width_chars(GTK_ENTRY(bitrate_entry), 8);
	gtk_entry_set_width_chars(GTK_ENTRY(entry.vorbis_quality_entry), 5);
	gtk_entry_set_editable(GTK_ENTRY(bitrate_entry), FALSE);
	gtk_entry_set_editable(GTK_ENTRY(entry.vorbis_quality_entry), FALSE);
	g_object_ref(item->frame); /* for multiple use */

	g_signal_connect(combo_list, "selection-changed", G_CALLBACK(vorbis_selection_changed_cb), bitrate_entry);
	g_signal_connect(entry.vorbis_mono_button, "toggled", G_CALLBACK(vorbis_mono_toggled_cb), NULL);
	g_signal_connect(entry.vorbis_stereo_button, "toggled", G_CALLBACK(vorbis_stereo_toggled_cb), NULL);

	gtk_widget_show_all(item->frame);
}

/********************************************************************/
static void create_faac_frame(void)
{
	tree_item_t *item;
	
	item = &codecs[2];
	item->frame = gtk_frame_new(item->label);
	g_object_ref(item->frame); /* for multiple use */
	gtk_widget_show_all(item->frame);
}

/********************************************************************/
static void vorbis_selection_changed_cb(GtkList *list, gpointer data)
{
	GtkWidget *bitrate_entry = data;
	gint quality;
	
	quality = atoi(gtk_entry_get_text(GTK_ENTRY(entry.vorbis_quality_entry)));
	
	switch (quality) {
	case 5:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~72kbps");
		break;
	case 10:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~80kbps");
		break;
	case 15:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~84kbps");
		break;
	case 20:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~89kbps");
		break;
	case 25:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~94kbps");
		break;
	case 30:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~103kbps");
		break;
	case 35:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~113kbps");
		break;
	case 40:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~126kbps");
		break;
	case 45:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~140kbps");
		break;
	case 50:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~154kbps");
		break;
	case 55:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~168kbps");
		break;
	case 60:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~182kbps");
		break;
	case 65:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~185kbps");
		break;
	case 70:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~209kbps");
		break;
	case 75:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~224kbps");
		break;
	case 80:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~244kbps");
		break;
	case 85:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~273kbps");
		break;
	case 90:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~316kbps");
		break;
	case 95:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~381kbps");
		break;
	case 100:
		gtk_entry_set_text(GTK_ENTRY(bitrate_entry), "~476kbps");
		break;
	}
}

/********************************************************************/
static void vorbis_mono_toggled_cb(GtkToggleButton *togglebutton, gpointer data)
{
	state.vorbis_channels = 1;
}

/********************************************************************/
static void vorbis_stereo_toggled_cb(GtkToggleButton *togglebutton, gpointer data)
{
	state.vorbis_channels = 2;
}

