/*
 * Copyright 2002 (C) Michal Zawalich <michuz@pld.org.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <vorbis/vorbisenc.h>
#include <sys/param.h>
#include "cdrom_handler.h"
#include "settings.h"
#include "vorbis_handler.h"

#define READ 1024
static gint bcount;     /* buffer counter */

static void add_comment(struct track_info *t, vorbis_comment *vc);
static void take_progress(struct track_info *t);
static void set_extension(struct track_info *t);

/********************************************************************/
void *gr_vorbis_encode(void *data)
{
	ogg_stream_state os; /* take physical pages, weld into a logical stream of packets */
	ogg_page og;         /* one Ogg bitstream page.  Vorbis packets are inside */
	ogg_packet op;       /* one raw packet of data for decode */
	vorbis_info vi;      /* struct that stores all the static vorbis bitstream settings */
	vorbis_comment vc;   /* struct that stores all the user comments */
	vorbis_dsp_state vd; /* central working state for the packet->PCM decoder */
	vorbis_block vb;     /* local working space for packet->PCM decode */
	signed char readbuffer[READ * 4 + 44]; /* out of the data segment, not the stack */
	int eos = 0;
	int i, founddata;
	FILE *wavfile;
	FILE *oggfile;
	struct track_info *t;

	/* thread mutex lock */
	g_mutex_lock(state.vorbis_mutex);

	/* some init stuff */
	set_extension(t = g_slist_nth_data(tracks_list, GPOINTER_TO_INT(data) - 1));
	wavfile = fopen(t->openfile, "rb");
	oggfile = fopen(t->savefile, "wb");
	readbuffer[0] = '\0';
	bcount = 0;

	/* we cheat on the WAV header; we just bypass the header and never
	 * verify that it matches 16bit/stereo/44.1kHz.  This is just an
	 * example, after all. */
	for (i = 0, founddata = 0; i < 30 && !feof(wavfile) && !ferror(wavfile); i++) {
		fread(readbuffer, 1, 2, wavfile);
		if (!strncmp(readbuffer, "da", 2)) {
			founddata = 1;
			fread(readbuffer, 1, 6, wavfile);
			break;
		}
	}
	
	/* ENCODE SETUP */
	/* choose an encoding mode */
	vorbis_info_init(&vi);
	vorbis_encode_init_vbr(&vi, state.vorbis_channels, 44100, (gfloat)state.vorbis_quality / 100);

	/* add a comment */
	vorbis_comment_init(&vc);
	add_comment(t, &vc);

	/* set up the analysis state and auxiliary encoding storage */
	vorbis_analysis_init(&vd, &vi);
	vorbis_block_init(&vd, &vb);

	/* set up our packet->stream encoder */
	/* pick a random serial number; that way we can more likely build
	 * chained streams just by concatenation */
	srand(time(NULL));
	ogg_stream_init(&os, rand());

	/* Vorbis streams begin with three headers; the initial header (with
	 * most of the codec setup parameters) which is mandated by the Ogg
	 * bitstream spec.  The second header holds any comment fields.  The
	 * third header holds the bitstream codebook.  We merely need to
	 * make the headers, then pass them to libvorbis one at a time;
	 * libvorbis handles the additional Ogg bitstream constraints */
	{
		ogg_packet header;
		ogg_packet header_comm;
		ogg_packet header_code;

		vorbis_analysis_headerout(&vd, &vc, &header, &header_comm, &header_code);
		ogg_stream_packetin(&os, &header); /* automatically placed in its own page */
		ogg_stream_packetin(&os, &header_comm);
		ogg_stream_packetin(&os, &header_code);

		/* We don't have to write out here, but doing so makes streaming 
		 * much easier, so we do, flushing ALL pages. This ensures the actual
		 * audio data will start on a new page */
		while (!eos) {
			int result = ogg_stream_flush(&os, &og);
			if (result == 0)
				break;
			fwrite(og.header, 1, og.header_len, oggfile);
			fwrite(og.body, 1, og.body_len, oggfile);
		}
	}

	while (!eos) {
		long i;
		long bytes = fread(readbuffer, 1, READ * 4, wavfile); /* stereo hardwired here */

		/* progress stuff */
		bcount++;
		take_progress(t);

		if (bytes == 0) {
			/* end of file.  this can be done implicitly in the mainline,
			 * but it's easier to see here in non-clever fashion.
			 * Tell the library we're at end of stream so that it can handle
			 * the last frame and mark end of stream in the output properly */
			vorbis_analysis_wrote(&vd, 0);
		} else {
			/* data to encode */
			/* expose the buffer to submit data */
			float **buffer = vorbis_analysis_buffer(&vd, READ);

			/* uninterleave samples */
			for (i = 0; i < bytes / 4; i++) {
				buffer[0][i] = ((readbuffer[i * 4 + 1] << 8) | (0x00ff & (int) readbuffer[i * 4])) / 32768.f;
				buffer[1][i] = ((readbuffer[i * 4 + 3] << 8) | (0x00ff & (int) readbuffer[i * 4 + 2])) / 32768.f;
			}
			/* tell the library how much we actually submitted */
			vorbis_analysis_wrote(&vd, i);
		}

		/* vorbis does some data preanalysis, then divvies up blocks for
		 * more involved (potentially parallel) processing.  Get a single
		 * block for encoding now */
		while (vorbis_analysis_blockout(&vd, &vb) == 1) {
			/* analysis, assume we want to use bitrate management */
			vorbis_analysis(&vb, NULL);
			vorbis_bitrate_addblock(&vb);
			while (vorbis_bitrate_flushpacket(&vd, &op)) {
				/* weld the packet into the bitstream */
				ogg_stream_packetin(&os, &op);
				/* write out pages (if any) */
				while (!eos) {
					int result = ogg_stream_pageout(&os, &og);
					if (result == 0)
						break;
					fwrite(og.header, 1, og.header_len, oggfile);
					fwrite(og.body, 1, og.body_len, oggfile);
					/* this could be set above, but for illustrative purposes, I do
					 * it here (to show that vorbis does know where the stream ends) */

					if (ogg_page_eos(&og))
						eos = 1;
				}
			}
		}
	}

	/* clean up and exit.  vorbis_info_clear() must be called last */
	ogg_stream_clear(&os);
	vorbis_block_clear(&vb);
	vorbis_dsp_clear(&vd);
	vorbis_comment_clear(&vc);
	vorbis_info_clear(&vi);

	fclose(wavfile);
	fclose(oggfile);
	unlink(t->openfile);

	state.ripsize += t->wavsize;
	
	/* thread mutex unlock */
	g_mutex_unlock(state.vorbis_mutex);
	pthread_exit(0);
	return (NULL);
}

/********************************************************************/
static void take_progress(struct track_info *t)
{
	gint readbytes = bcount * (READ * 4 + 44);

	state.progress = ((gfloat) readbytes / (gfloat) t->wavsize) * 100;
	state.progress2 = ((gfloat) (state.ripsize + readbytes) / (gfloat) state.totalsize) * 100;

	g_free(state.progress_text);
	g_free(state.progress_label1);
	g_free(state.progress_label2);

	state.progress_text = g_strdup_printf("%d%%", state.progress);
	state.progress_label1 = g_strdup_printf("INPUT: %s", g_strrstr(t->openfile, "/") + 1);
	state.progress_label2 = g_strdup_printf("OUTPUT: %s", g_strrstr(t->savefile, "/") + 1);
}

/********************************************************************/
static void add_comment(struct track_info *t, vorbis_comment *vc)
{
	vorbis_comment_add_tag(vc, "ENCODER", PACKAGE " v" VERSION " Copyright(c) Michal Zawalich");
	vorbis_comment_add_tag(vc, "TITLE", t->title);
	vorbis_comment_add_tag(vc, "ARTIST", t->artist);
	vorbis_comment_add_tag(vc, "ALBUM", t->album);
	vorbis_comment_add_tag(vc, "DATE", t->year);
	vorbis_comment_add_tag(vc, "GENRE", t->genre);
	vorbis_comment_add_tag(vc, "DESCRIPTION", t->comment);
}

/********************************************************************/
static void set_extension(struct track_info *t)
{
	char outpath[MAXPATHLEN];
	char *tmp;
		
	strcpy(outpath, t->openfile);
	tmp = g_strrstr(outpath, "wav");
	strcpy(tmp, "ogg");
	g_free(t->savefile);
	t->savefile = g_strdup(outpath);
}
